@@ -74,26 +74,26 @@ FORMAT_PRECEDENCE = ["s", "b", "i", "a", "c"]
 
 
 def _extract_text_and_format_from_ast(item: dict):
-    type = item["type"]
+    literal = item.get("literal", "")
+    item_type = item["type"]
 
-    if type == "html_inline":
-        if item.get("literal", "") == "<s>":
+    if item_type == "html_inline":
+        if literal == "<s>":
             return "", ("s",)
 
-    if type == "emph":
-        return item.get("literal", ""), ("i",)
+    if item_type == "strong":
+        return literal, ("b",)
 
-    if type == "strong":
-        return item.get("literal", ""), ("b",)
+    if item_type == "emph":
+        return literal, ("i",)
 
-    if type == "code":
-        return item.get("literal", ""), ("c",)
+    if item_type == "link":
+        return literal, ("a", item.get("destination", ""))
 
-    if type == "link":
-        if "destination" in item:
-            return item.get("literal", ""), ("a", item["destination"])
+    if item_type == "code":
+        return literal, ("c",)
 
-    return item.get("literal", ""), ()
+    return literal, ()
 
 
 def _get_format(notion_segment, as_set=False):
@@ -137,7 +137,8 @@ def markdown_to_notion(markdown: str) -> list:
         Blocks converted from input.
     """
 
-    # commonmark doesn't support strikethrough, so we need to handle it ourselves
+    # commonmark doesn't support strikethrough,
+    # so we need to handle it ourselves
     while markdown.count("~~") >= 2:
         markdown = markdown.replace("~~", "<s>", 1)
         markdown = markdown.replace("~~", "</s>", 1)
@@ -182,7 +183,8 @@ def markdown_to_notion(markdown: str) -> list:
                     else [literal]
                 )
 
-            # in the ast format, code blocks are meant to be immediately self-closing
+            # in the ast format, code blocks are meant
+            # to be immediately self-closing
             if ("c",) in format:
                 format.remove(("c",))
 
@@ -329,8 +331,8 @@ def notion_to_plaintext(notion: list, client=None) -> str:
     Arguments
     ---------
     notion : list
-        List of Notion Blocks
-        TODO: is it true?
+        Text in a Notion specific API format
+        i.e. [["some text"]]
 
     client : NotionClient, optional
         Used for getting blocks, if passed.
