@@ -1,18 +1,20 @@
 import json
-import threading
+from threading import Thread
 import uuid
-from collections import defaultdict, Callable
+from collections import defaultdict
+from typing import Callable
 from copy import deepcopy
 from inspect import signature
 from pathlib import Path
 from threading import Lock
+from typing import Union
 
 from dictdiffer import diff
 from tzlocal import get_localzone
 
 from notion.logger import logger
 from notion.settings import NOTION_CACHE_DIR
-from notion.utils import extract_id
+from notion.utils import extract_id, to_list
 
 
 class MissingClass:
@@ -40,37 +42,43 @@ class Callback:
             difference, old_val, new_val
         )
 
-        logger.debug("Firing callback {} with kwargs: {}".format(self.callback, kwargs))
+        logger.debug(f"Firing callback {self.callback} with kwargs: {kwargs}")
 
-        # trim down the parameters we'll be passing, to include only those the callback will accept
+        # trim down the passed parameters
+        # to include only those the callback will accept
         params = signature(self.callback).parameters
-        if not any(["**" in str(param) for param in params.values()]):
-            # there's no "**kwargs" in the callback signature, so remove any unaccepted params
-            for arg in list(kwargs.keys()):
+        if not any(["**" in str(p) for p in params.values()]):
+            # there's no "**kwargs" in the callback signature,
+            # so remove any unaccepted params
+            for arg in kwargs.keys():
                 if arg not in params:
                     del kwargs[arg]
 
         # perform the callback, gracefully handling any exceptions
         try:
-            # trigger the callback within its own thread, so it won't block others if it's long-running
-            threading.Thread(target=self.callback, kwargs=kwargs, daemon=True).start()
+            # trigger the callback within its own thread,
+            # so it won't block others if it's long-running
+            Thread(target=self.callback, kwargs=kwargs, daemon=True).start()
         except Exception as e:
             logger.error(
-                "Error while processing callback for {}: {}".format(
-                    repr(self.record), repr(e)
-                )
+                f"Error while processing callback for {repr(self.record)}: {repr(e)}"
             )
 
-    def __eq__(self, val):
-        if isinstance(val, str):
-            return self.callback_id.startswith(val)
-        elif isinstance(val, Callback):
-            return self.callback_id == val.callback_id
-        else:
-            return False
+    def __eq__(self, value: Union["Callback", str]) -> bool:
+        if isinstance(value, str):
+            return self.callback_id.startswith(value)
+
+        if isinstance(value, Callback):
+            return self.callback_id == value.callback_id
+
+        return False
 
 
-class RecordStore(object):
+class RecordStore:
+    """
+    Central Record Store.
+    """
+
     def __init__(self, client, cache_key=None):
         self._mutex = Lock()
         self._client = client
@@ -84,91 +92,68 @@ class RecordStore(object):
         with self._mutex:
             self._load_cache()
 
-    def _get(self, table, id):
-        return self._values[table].get(id, Missing)
-
-    def add_callback(self, record, callback, callback_id=None, **extra_kwargs):
-        if not callable(callback):
-            raise ValueError(
-                f"The callback {callback} must be a 'callable' object, such as a function."
-            )
-
-        self.remove_callbacks(record._table, record.id, callback_id)
-        callback_obj = Callback(
-            callback, record, callback_id=callback_id, extra_kwargs=extra_kwargs
-        )
-        self._callbacks[record._table][record.id].append(callback_obj)
-        return callback_obj
-
-    def remove_callbacks(self, table, id, callback_or_callback_id_prefix=""):
-        """
-        Remove all callbacks for the record specified by `table` and `id` that have a callback_id
-        starting with the string `callback_or_callback_id_prefix`, or are equal to the provided callback.
-        """
-        if callback_or_callback_id_prefix is None:
-            return
-        callbacks = self._callbacks[table][id]
-        while callback_or_callback_id_prefix in callbacks:
-            callbacks.remove(callback_or_callback_id_prefix)
+    def _get(self, table: str, record_id: str):
+        return self._values[table].get(record_id, Missing)
 
     def _get_cache_path(self, attribute):
-        return str(Path(NOTION_CACHE_DIR) / f"{self._cache_key}{attribute}.json")
+        file = f"{self._cache_key}{attribute}.json"
+        return str(Path(NOTION_CACHE_DIR) / file)
 
     def _load_cache(self, attributes=("_values", "_role", "_collection_row_ids")):
         if not self._cache_key:
             return
+
         for attr in attributes:
             try:
                 with open(self._get_cache_path(attr)) as f:
                     if attr == "_collection_row_ids":
                         self._collection_row_ids.update(json.load(f))
+
                     else:
                         for k, v in json.load(f).items():
                             getattr(self, attr)[k].update(v)
+
             except (FileNotFoundError, ValueError):
                 pass
 
-    def set_collection_rows(self, collection_id, row_ids):
+    def _save_cache(self, attribute):
+        if not self._cache_key:
+            return
 
+        with open(self._get_cache_path(attribute), "w") as f:
+            json.dump(getattr(self, attribute), f)
+
+    def _trigger_callbacks(self, table, record_id, difference, old_val, new_val):
+        for callback_obj in self._callbacks[table][record_id]:
+            callback_obj(difference, old_val, new_val)
+
+    def set_collection_rows(self, collection_id: str, row_ids):
         if collection_id in self._collection_row_ids:
-            old_ids = set(self._collection_row_ids[collection_id])
+            old_ids = set(self.get_collection_rows(collection_id))
             new_ids = set(row_ids)
-            added = new_ids - old_ids
-            removed = old_ids - new_ids
-            for id in added:
-                self._trigger_callbacks(
-                    "collection",
-                    collection_id,
-                    [("row_added", "rows", id)],
-                    old_ids,
-                    new_ids,
-                )
-            for id in removed:
-                self._trigger_callbacks(
-                    "collection",
-                    collection_id,
-                    [("row_removed", "rows", id)],
-                    old_ids,
-                    new_ids,
-                )
+            args = {
+                "table": "collection",
+                "record_id": collection_id,
+                "old_val": old_ids,
+                "new_val": new_ids,
+            }
+
+            for i in new_ids - old_ids:
+                args["difference"] = [("row_added", "rows", i)]
+                self._trigger_callbacks(**args)
+
+            for i in old_ids - new_ids:
+                args["difference"] = [("row_removed", "rows", i)]
+                self._trigger_callbacks(**args)
+
         self._collection_row_ids[collection_id] = row_ids
         self._save_cache("_collection_row_ids")
 
     def get_collection_rows(self, collection_id):
         return self._collection_row_ids.get(collection_id, [])
 
-    def _save_cache(self, attribute):
-        if not self._cache_key:
-            return
-        with open(self._get_cache_path(attribute), "w") as f:
-            json.dump(getattr(self, attribute), f)
-
-    def _trigger_callbacks(self, table, id, difference, old_val, new_val):
-        for callback_obj in self._callbacks[table][id]:
-            callback_obj(difference, old_val, new_val)
-
-    def get_role(self, table, id, force_refresh=False):
-        self.get(table, id, force_refresh=force_refresh)
+    def get_role(self, table, record_id, force_refresh=False):
+        self.get(table, record_id, force_refresh=force_refresh)
         return self._role[table].get(id, None)
 
     def get(self, table, url_or_id, force_refresh=False):
@@ -184,20 +169,47 @@ class RecordStore(object):
             result = self._get(table, rid)
         return result if result is not Missing else None
 
-    def _update_record(self, table, id, value=None, role=None):
+    def add_callback(
+        self, record, callback: Callable, callback_id=None, **extra_kwargs
+    ):
+        if not callable(callback):
+            raise ValueError(f"The callback {callback} must be a callable.")
 
+        self.remove_callbacks(record._table, record.id, callback_id)
+        callback_obj = Callback(
+            callback, record, callback_id=callback_id, extra_kwargs=extra_kwargs
+        )
+        self._callbacks[record._table][record.id].append(callback_obj)
+        return callback_obj
+
+    def remove_callbacks(self, table, record_id: str, cb_or_cb_id_prefix=""):
+        """
+        Remove all callbacks for the record specified
+        by `table` and `id` that have a callback_id
+        starting with the string `cb_or_cb_id_prefix`,
+        or are equal to the provided callback.
+        """
+        if cb_or_cb_id_prefix is None:
+            return
+
+        callbacks = self._callbacks[table][record_id]
+        while cb_or_cb_id_prefix in callbacks:
+            callbacks.remove(cb_or_cb_id_prefix)
+
+    def _update_record(self, table, record_id, value=None, role=None):
         callback_queue = []
 
         with self._mutex:
             if role:
-                logger.debug("Updating 'role' for {}/{} to {}".format(table, id, role))
-                self._role[table][id] = role
+                logger.debug(f"Updating 'role' for '{table}/{record_id}' to '{role}'")
+                self._role[table][record_id] = role
                 self._save_cache("_role")
             if value:
+                p_value = json.dumps(value, indent=2)
                 logger.debug(
-                    "Updating 'value' for {}/{} to {}".format(table, id, value)
+                    f"Updating 'value' for '{table}/{record_id}' to \n{p_value}"
                 )
-                old_val = self._values[table][id]
+                old_val = self._values[table][record_id]
                 difference = list(
                     diff(
                         old_val,
@@ -206,11 +218,13 @@ class RecordStore(object):
                         expand=True,
                     )
                 )
-                self._values[table][id] = value
+                self._values[table][record_id] = value
                 self._save_cache("_values")
                 if old_val and difference:
-                    logger.debug("Value changed! Difference: {}".format(difference))
-                    callback_queue.append((table, id, difference, old_val, value))
+                    p_difference = json.dumps(value, indent=2)
+                    logger.debug(f"Value changed! Difference:\n{p_difference}")
+                    callback = (table, record_id, difference, old_val, value)
+                    callback_queue.append(callback)
 
         # run callbacks outside the mutex to avoid lockups
         for cb in callback_queue:
@@ -218,56 +232,52 @@ class RecordStore(object):
 
     def call_get_record_values(self, **kwargs):
         """
-        Call the server's getRecordValues endpoint to update the local record store. The keyword arguments map
-        table names into lists of (or singular) record IDs to load for that table. Use True to refresh all known
-        records for that table.
+        Call the server's getRecordValues endpoint
+        to update the local record store.
+        The keyword arguments map table names into lists
+        of (or singular) record IDs to load for that table.
+        Use True to refresh all known records for that table.
         """
-
-        requestlist = []
+        requests = []
 
         for table, ids in kwargs.items():
 
             # ensure "ids" is a proper list
-            if ids is True:
+            if ids:
                 ids = list(self._values.get(table, {}).keys())
-            if isinstance(ids, str):
-                ids = [ids]
+            ids = to_list(ids)
 
-            # if we're in a transaction, add the requested IDs to a queue to refresh when the transaction completes
+            # if we're in a transaction, add the requested IDs
+            # to a queue to refresh when the transaction completes
             if self._client.in_transaction():
-                self._records_to_refresh[table] = list(
-                    set(self._records_to_refresh.get(table, []) + ids)
-                )
+                records = set(self._records_to_refresh.get(table, []) + ids)
+                self._records_to_refresh[table] = list(records)
                 continue
 
-            requestlist += [{"table": table, "id": extract_id(id)} for id in ids]
+            requests += [{"table": table, "id": extract_id(i)} for i in ids]
 
-        if requestlist:
-            logger.debug(
-                "Calling 'getRecordValues' endpoint for requests: {}".format(
-                    requestlist
-                )
-            )
-            results = self._client.post(
-                "getRecordValues", {"requests": requestlist}
-            ).json()["results"]
-            for request, result in zip(requestlist, results):
+        if requests:
+            logger.debug(f"Calling 'getRecordValues' endpoint for requests: {requests}")
+            data = {"requests": requests}
+            data = self._client.post("getRecordValues", data).json()
+            results = data["results"]
+
+            for request, result in zip(requests, results):
                 self._update_record(
-                    request["table"],
-                    request["id"],
+                    table=request["table"],
+                    record_id=request["id"],
                     value=result.get("value"),
                     role=result.get("role"),
                 )
 
-    def get_current_version(self, table, id):
-        values = self._get(table, id)
+    def get_current_version(self, table, record_id):
+        values = self._get(table, record_id)
         if values and "version" in values:
             return values["version"]
-        else:
-            return -1
+
+        return -1
 
     def call_load_page_chunk(self, page_id):
-
         if self._client.in_transaction():
             self._pages_to_refresh.append(page_id)
             return
@@ -279,18 +289,20 @@ class RecordStore(object):
             "chunkNumber": 0,
             "verticalColumns": False,
         }
+        data = self._client.post("loadPageChunk", data).json()
+        self.store_record_map(data)
 
-        recordmap = self._client.post("loadPageChunk", data).json()["recordMap"]
-
-        self.store_recordmap(recordmap)
-
-    def store_recordmap(self, recordmap):
-        # TODO: accept response dict and fetch key recordMap here?
-        for table, records in recordmap.items():
-            for id, record in records.items():
+    def store_record_map(self, data: dict) -> dict:
+        data = data["recordMap"]
+        for table, records in data.items():
+            for record_id, record in records.items():
                 self._update_record(
-                    table, id, value=record.get("value"), role=record.get("role")
+                    table=table,
+                    record_id=record_id,
+                    value=record.get("value"),
+                    role=record.get("role"),
                 )
+        return data
 
     def call_query_collection(
         self,
@@ -301,6 +313,7 @@ class RecordStore(object):
         aggregate: list = None,
         aggregations: list = None,
         filter: dict = None,
+        filter_operator: str = "and",
         sort: list = [],
         calendar_by: str = "",
         group_by: str = "",
@@ -309,19 +322,13 @@ class RecordStore(object):
 
         if aggregate and aggregations:
             raise ValueError(
-                "Use only one of `aggregate` or `aggregations` (old vs new format)"
+                "Use either `aggregate` or `aggregations` (old vs new format)"
             )
 
-        aggregate = aggregate or []
+        aggregate = to_list(aggregate or [])
         aggregations = aggregations or []
-        filter = filter or {}
-        sort = sort or []
-
-        # convert singletons into lists if needed
-        if isinstance(aggregate, dict):
-            aggregate = [aggregate]
-        if isinstance(sort, dict):
-            sort = [sort]
+        filter = to_list(filter or {})
+        sort = to_list(sort or [])
 
         data = {
             "collectionId": collection_id,
@@ -337,20 +344,19 @@ class RecordStore(object):
             "query": {
                 "aggregate": aggregate,
                 "aggregations": aggregations,
-                "filter": filter,
+                "filter": {
+                    "filters": filter,
+                    "filter_operator": filter_operator,
+                },
                 "sort": sort,
             },
         }
+        data = self._client.post("queryCollection", data).json()
+        self.store_record_map(data)
 
-        response = self._client.post("queryCollection", data).json()
-
-        self.store_recordmap(response["recordMap"])
-
-        # TODO: dict or list?
-        return response["result"]
+        return data["result"]
 
     def handle_post_transaction_refreshing(self):
-
         for block_id in self._pages_to_refresh:
             self.call_load_page_chunk(block_id)
         self._pages_to_refresh = []
@@ -358,21 +364,10 @@ class RecordStore(object):
         self.call_get_record_values(**self._records_to_refresh)
         self._records_to_refresh = {}
 
-    def run_local_operations(self, operations):
-        """
-        Called to simulate the results of running the operations
-        on the server, to keep the record store in sync even when
-        we haven't completed a refresh (or we did a refresh but
-        the database hadn't actually updated yet...)
-        """
-        for operation in operations:
-            self.run_local_operation(**operation)
-
-    def run_local_operation(self, table, id, path, command, args):
-
+    def run_local_operation(self, table, record_id, path, command, args):
         with self._mutex:
             path = deepcopy(path)
-            new_val = deepcopy(self._values[table][id])
+            new_val = deepcopy(self._values[table][record_id])
 
         ref = new_val
 
@@ -389,27 +384,31 @@ class RecordStore(object):
 
         if command == "update":
             ref.update(args)
-        elif command == "set":
+
+        if command == "set":
             if path:
                 ref[path[0]] = args
             else:
-                # this is the case of "setting the top level" (i.e. creating a record)
+                # case for "setting the top level" (i.e. creating a record)
                 ref.clear()
                 ref.update(args)
-        elif command == "listAfter":
+
+        if command == "listAfter":
             if "after" in args:
                 ref.insert(ref.index(args["after"]) + 1, args["id"])
             else:
                 ref.append(args["id"])
-        elif command == "listBefore":
+
+        if command == "listBefore":
             if "before" in args:
                 ref.insert(ref.index(args["before"]), args["id"])
             else:
                 ref.insert(0, args["id"])
-        elif command == "listRemove":
+
+        if command == "listRemove":
             try:
                 ref.remove(args["id"])
             except ValueError:
                 pass
 
-        self._update_record(table, id, value=new_val)
+        self._update_record(table, record_id, value=new_val)
