@@ -1,27 +1,36 @@
+import os
 import uuid
 from datetime import datetime
-from typing import Union, Iterable, Any
+from typing import Any, Optional
 from urllib.parse import urlparse, parse_qs, quote_plus, unquote_plus
 
-import requests
-from bs4 import BeautifulSoup
 from slugify import slugify as _dash_slugify
 
 from notion.settings import (
     BASE_URL,
     SIGNED_URL_PREFIX,
     S3_URL_PREFIX,
-    S3_URL_PREFIX_ENCODED,
     EMBED_API_URL,
 )
 
 
-class InvalidNotionIdentifier(Exception):
-    """
-    Invalid Notion Identifier was found.
+def to_list(value) -> list:
     """
+    Wrap value in list if it's not already in a list.
 
-    pass
+
+    Arguments
+    ---------
+    value : Any
+        Value to wrap in list.
+
+
+    Returns
+    -------
+    list
+        List with value inside.
+    """
+    return value if isinstance(value, list) else [value]
 
 
 def now() -> int:
@@ -37,38 +46,65 @@ def now() -> int:
     return int(datetime.now().timestamp() * 1000)
 
 
-def extract_id(url_or_id: str) -> str:
+def human_size(path: str, divider: int = 1024) -> str:
     """
-    Extract the block/page ID from a Notion.so URL.
-
-    If it's a bare page URL, it will be the ID of the page.
-    If there's a hash with a block ID in it (from clicking "Copy Link")
-    on a block in a page), it will instead be the ID of that block.
-    If it's already in ID format, it will be passed right through.
+    Get human readable file size.
 
 
     Arguments
     ---------
-    url_or_id : str
-        Link to block or its ID.
+    path : str
+        Path to the file.
 
-
-    Raises
-    ------
-    InvalidNotionIdentifier
-        Raised when `url_or_id` can't be converted to UUID.
+    divider : int, optional
+        Divider used for calculations, use 1000 or 1024.
+        Defaults to 1024.
 
 
     Returns
     -------
     str
-        ID of the block.
+        Converted size.
     """
-    original_url_or_id = url_or_id
+    size, divider = os.path.getsize(path), float(divider)
+    size = size / divider if size < divider else size
 
-    if url_or_id.startswith(BASE_URL):
-        url_or_id = (
-            url_or_id.split("#")[-1]
+    for unit in ("KB", "KB", "MB", "GB", "TB"):
+        if abs(size) < divider:
+            return f"{size:.1f}{unit}"
+        size /= divider
+
+    return str(size)
+
+
+def extract_id(source) -> Optional[str]:
+    """
+    Extract the record ID from a block or Notion.so URL.
+
+    If it's a bare page URL, it will be the ID of the page.
+    If there's a hash with a block ID in it (from clicking "Copy Link")
+    on a block in a page), it will instead be the ID of that block.
+    If it's already in ID format, it will be passed right through.
+    If it's a Block, it will be the ID of a block.
+
+
+    Arguments
+    ---------
+    source : Block or str
+        Block or Link to block or its ID.
+
+
+    Returns
+    -------
+    str
+        ID of the block or None.
+    """
+    if not isinstance(source, str):
+        return source.id
+
+    if source.startswith(BASE_URL):
+        source = (
+            source.split("#")[-1]
             .split("/")[-1]
             .split("&p=")[-1]
             .split("?")[0]
@@ -76,31 +112,12 @@ def extract_id(url_or_id: str) -> str:
         )
 
     try:
-        return str(uuid.UUID(url_or_id))
+        return str(uuid.UUID(source))
     except ValueError:
-        raise InvalidNotionIdentifier(original_url_or_id)
+        return None
 
 
-def get_embed_data(source_url: str) -> dict:
-    """
-    Get embed data.
-
-
-    Arguments
-    ---------
-    source_url : str
-        Source URL from which the embedded data will be extracted.
-
-
-    Returns
-    -------
-    dict
-        Extracted data.
-    """
-    return requests.get(f"{EMBED_API_URL}&url={source_url}").json()
-
-
-def get_embed_link(source_url: str) -> str:
+def get_embed_link(source_url: str, client) -> str:
     """
     Get embed link.
 
@@ -110,21 +127,21 @@ def get_embed_link(source_url: str) -> str:
     source_url : str
         Source URL from which the embedded link will be extracted.
 
+    client : NotionClient
+        Client used for sending the actual request.
+
 
     Returns
     -------
     str
         Extracted link.
     """
-
-    data = get_embed_data(source_url)
+    data = client.get(f"{EMBED_API_URL}&url={source_url}").json()
 
     if "html" not in data:
         return source_url
 
-    # TODO: replace BS4 with built in solution
-    url = list(BeautifulSoup(data["html"], "html.parser").children)[0]["src"]
-
+    url = data["html"].split('src="')[1].split('"')[0]
     return parse_qs(urlparse(url).query)["src"][0]
 
 
@@ -139,7 +156,8 @@ def add_signed_prefix_as_needed(url: str, client=None) -> str:
         URL to operate on.
 
     client : NotionClient, optional
-        # TODO: Client object, used for ???
+        It's used for making wrapped requests via
+        initialized requests.Session object.
         Defaults to None.
 
 
@@ -152,7 +170,9 @@ def add_signed_prefix_as_needed(url: str, client=None) -> str:
         return ""
 
     if url.startswith(S3_URL_PREFIX):
-        url = SIGNED_URL_PREFIX + quote_plus(url)
+        path, query = url.split("?")
+        url = f"{SIGNED_URL_PREFIX}{quote_plus(path)}?{query}"
+
         if client:
             url = client.session.head(url).headers.get("Location")
 
@@ -175,27 +195,20 @@ def remove_signed_prefix_as_needed(url: str) -> str:
     str
         Non-prefixed URL.
     """
-    if not url:
-        return ""
-
     if url.startswith(SIGNED_URL_PREFIX):
-        return unquote_plus(url[len(S3_URL_PREFIX) :])
+        url = unquote_plus(url[len(S3_URL_PREFIX) :])
 
-    if url.startswith(S3_URL_PREFIX_ENCODED):
-        parsed = urlparse(url.replace(S3_URL_PREFIX_ENCODED, S3_URL_PREFIX))
-        return f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
-
-    return url
+    return url or ""
 
 
-def slugify(original: str) -> str:
+def slugify(text: str) -> str:
     """
     Convert text to computer-friendly simplified form.
 
 
     Arguments
     ---------
-    original : str
+    text : str
         String to operate on.
 
 
@@ -204,18 +217,18 @@ def slugify(original: str) -> str:
     str
         Converted string.
     """
-    return _dash_slugify(original).replace("-", "_")
+    return _dash_slugify(text).replace("-", "_")
 
 
-def get_by_path(path: Union[Iterable, str], obj: Any, default: Any = None):
+def get_by_path(path: str, obj: Any, default: Any = None) -> dict:
     """
     Get value from object's key by dotted path (i.e. "path.to.some.key").
 
 
     Arguments
     ---------
-    path : list or str
-        Path in string form or as list elements.
+    path : str
+        Path in string form.
 
     obj : Any
         Object to traverse.
@@ -227,21 +240,19 @@ def get_by_path(path: Union[Iterable, str], obj: Any, default: Any = None):
 
     Returns
     -------
-    Value stored under specified key or default value.
+    dict
+        Value stored under specified key or default value.
     """
-    if isinstance(path, str):
-        path = path.split(".")
-
-    value = obj
+    path = path.split(".")
 
     # try to traverse down the sequence of keys defined
     # in the path, to get the target value if it exists
     try:
         for key in path:
-            if isinstance(value, list):
+            if isinstance(obj, list):
                 key = int(key)
-            value = value[key]
+            obj = obj[key]
     except (KeyError, TypeError, IndexError):
-        value = default
+        obj = default
 
-    return value
+    return obj
